//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Tuning/ParameterTuningWidget.cpp
//! @brief     Implements class ParameterTuningWidget.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Tuning/ParameterTuningWidget.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Data/Data2DItem.h"
#include "GUI/Model/Job/ParameterTreeItems.h"
#include "GUI/Model/Model/ParameterTuningModel.h"
#include "GUI/Model/Project/ProjectDocument.h"
#include "GUI/View/Info/CautionSign.h"
#include "GUI/View/ParEdit/ParameterTuningDelegate.h"
#include "GUI/View/Tool/mainwindow_constants.h"
#include "GUI/View/Tuning/SliderSettingsWidget.h"
#include "GUI/View/Widget/StyledToolbar.h"
#include <QAction>
#include <QHeaderView>
#include <QSettings>
#include <QTreeView>
#include <QVBoxLayout>

namespace {
void expandChildren(const QModelIndex& parent, QTreeView* view)
{
    int childCount = view->model()->rowCount(parent);
    for (int i = 0; i < childCount; i++) {
        const QModelIndex& child = view->model()->index(i, 0, parent);
        expandChildren(child, view);
    }

    if (view->model()->data(parent, Qt::UserRole).toBool())
        view->collapse(parent);
    else
        view->expand(parent);
}

int currentColumnWidth = GUI::Style::PARAMETER_TUNING_WIDGET_COLUMN_WIDTH;
} // namespace

ParameterTuningWidget::ParameterTuningWidget(QWidget* parent)
    : DataAccessWidget(parent)
    , m_jobModel(nullptr)
    , m_parameterTuningModel(nullptr)
    , m_sliderSettingsWidget(new SliderSettingsWidget(this))
    , m_treeView(new QTreeView)
    , m_delegate(new ParameterTuningDelegate(this))
    , m_cautionSign(new CautionSign(m_treeView))
{
    setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);

    m_treeView->setItemDelegate(m_delegate); // the delegate creates view&control for one param
    m_treeView->setContextMenuPolicy(Qt::CustomContextMenu);
    m_treeView->setDragDropMode(QAbstractItemView::NoDragDrop);
    m_treeView->setAttribute(Qt::WA_MacShowFocusRect, false);

    auto* resetValuesAction = new QAction(QIcon(":/images/undo.svg"), "Reset values", this);
    resetValuesAction->setToolTip("Reset parameter tree to initial values");
    connect(resetValuesAction, &QAction::triggered, this,
            &ParameterTuningWidget::restoreModelsOfCurrentJobItem);

    auto* toolbar = new StyledToolbar(this);
    toolbar->setToolButtonStyle(Qt::ToolButtonTextBesideIcon);
    toolbar->addAction(resetValuesAction);

    auto* mainLayout = new QVBoxLayout;
    mainLayout->setContentsMargins(0, 0, 0, 0);
    mainLayout->setSpacing(0);
    mainLayout->addWidget(toolbar);
    mainLayout->addWidget(m_sliderSettingsWidget);
    mainLayout->addWidget(m_treeView);
    setLayout(mainLayout);

    connect(m_sliderSettingsWidget, &SliderSettingsWidget::sliderRangeFactorChanged, this,
            &ParameterTuningWidget::onSliderRangeChanged);
    connect(m_sliderSettingsWidget, &SliderSettingsWidget::lockzChanged, this,
            &ParameterTuningWidget::onLockZValueChanged);
    connect(m_delegate, &ParameterTuningDelegate::currentLinkChanged, this,
            &ParameterTuningWidget::onCurrentLinkChanged);
    connect(m_treeView, &QTreeView::customContextMenuRequested, this,
            &ParameterTuningWidget::onCustomContextMenuRequested);

    applySettings();
}

ParameterTuningWidget::~ParameterTuningWidget()
{
    saveSettings();
}

void ParameterTuningWidget::saveSettings()
{
    QSettings settings;
    settings.beginGroup(GUI::Style::S_PARAMETER_TUNING_WIDGET);
    settings.setValue(GUI::Style::S_PARAMETER_TUNING_WIDGET_COLUMN_WIDTH, ::currentColumnWidth);
    settings.endGroup();
    settings.sync();
}

void ParameterTuningWidget::applySettings()
{
    QSettings settings;
    if (settings.childGroups().contains(GUI::Style::S_PARAMETER_TUNING_WIDGET)) {
        settings.beginGroup(GUI::Style::S_PARAMETER_TUNING_WIDGET);
        ::currentColumnWidth =
            settings.value(GUI::Style::S_PARAMETER_TUNING_WIDGET_COLUMN_WIDTH).toInt();
        settings.endGroup();
    }
}

void ParameterTuningWidget::setJobOrDatafileItem(QObject* job_item)
{
    setBaseItem(job_item);
    m_sliderSettingsWidget->setJobOrDatafileItem(job_item);

    updateParameterModel();
    updateDragAndDropSettings();

    connect(jobItem(), &JobItem::jobStatusChanged, this,
            [this](const JobStatus) { updateJobStatus(); });

    updateJobStatus();
}

void ParameterTuningWidget::setModel(QObject* jobModel)
{
    m_jobModel = dynamic_cast<JobModel*>(jobModel);
    ASSERT(m_jobModel);
}

QItemSelectionModel* ParameterTuningWidget::selectionModel()
{
    ASSERT(m_treeView);
    return m_treeView->selectionModel();
}

//! Returns list of ParameterItem's currently selected in parameter tree

QVector<ParameterItem*> ParameterTuningWidget::selectedParameterItems()
{
    QVector<ParameterItem*> result;
    for (auto index : selectionModel()->selectedIndexes())
        if (ParameterItem* parItem = m_parameterTuningModel->getParameterItem(index))
            result.push_back(parItem);

    return result;
}

void ParameterTuningWidget::onCurrentLinkChanged(ParameterItem* item)
{
    ASSERT(jobItem());

    if (jobItem()->isRunning())
        return;

    if (item)
        m_jobModel->runJob(jobItem());
}

void ParameterTuningWidget::onSliderRangeChanged(int value)
{
    m_delegate->setSliderRangeFactor(value);
}

void ParameterTuningWidget::onLockZValueChanged(bool value)
{
    if (!jobItem())
        return;
    if (Data2DItem* data2DItem = jobItem()->data2DItem())
        data2DItem->setZaxisLocked(value);
}

void ParameterTuningWidget::updateParameterModel()
{
    ASSERT(m_jobModel);

    if (!jobItem())
        return;

    if (!jobItem()->sampleItem() || !jobItem()->instrumentItem())
        throw std::runtime_error("JobItem is missing sample or instrument model");

    delete m_parameterTuningModel;
    m_parameterTuningModel =
        new ParameterTuningModel(jobItem()->parameterContainerItem()->parameterTreeRoot(), this);

    m_treeView->setModel(m_parameterTuningModel);
    m_treeView->setColumnWidth(0, ::currentColumnWidth);
    expandChildren(m_treeView->rootIndex(), m_treeView);

    connect(m_treeView, &QTreeView::expanded, m_parameterTuningModel,
            &ParameterTuningModel::setExpanded);
    connect(m_treeView, &QTreeView::collapsed, m_parameterTuningModel,
            &ParameterTuningModel::setCollapsed);

    connect(m_treeView->header(), &QHeaderView::sectionResized, [](int i, int, int n) {
        if (i == 0)
            ::currentColumnWidth = n;
    });
}

void ParameterTuningWidget::onCustomContextMenuRequested(const QPoint& point)
{
    emit itemContextMenuRequest(m_treeView->mapToGlobal(point + QPoint(2, 22)));
}

void ParameterTuningWidget::restoreModelsOfCurrentJobItem()
{
    ASSERT(m_jobModel);
    ASSERT(jobItem());

    if (jobItem()->isRunning())
        return;

    closeActiveEditors();

    m_jobModel->restore(jobItem());
    m_jobModel->runJob(jobItem());
    gProjectDocument.value()->setModified();
}

void ParameterTuningWidget::makeSelected(ParameterItem* item)
{
    QModelIndex index = m_parameterTuningModel->indexForItem(item);
    if (index.isValid())
        selectionModel()->select(index, QItemSelectionModel::Select);
}

void ParameterTuningWidget::contextMenuEvent(QContextMenuEvent*)
{
    // reimplemented to suppress context menu from QMainWindow
}

//! Disable drag-and-drop abilities, if job is in fit running state.

void ParameterTuningWidget::updateDragAndDropSettings()
{
    ASSERT(jobItem());
    if (jobItem()->status() == JobStatus::Fitting) {
        setTuningDelegateEnabled(false);
        m_treeView->setDragDropMode(QAbstractItemView::NoDragDrop);
    } else {
        setTuningDelegateEnabled(true);
        if (jobItem()->isValidForFitting())
            m_treeView->setDragDropMode(QAbstractItemView::DragOnly);
    }
}

//! Sets delegate to enabled/disabled state.
//! In 'disabled' state the delegate is in ReadOnlyMode, if it was containing already some
//! editing widget, it will be forced to close.
void ParameterTuningWidget::setTuningDelegateEnabled(bool enabled)
{
    if (enabled)
        m_delegate->setReadOnly(false);
    else {
        m_delegate->setReadOnly(true);
        closeActiveEditors();
    }
}

void ParameterTuningWidget::closeActiveEditors()
{
    QModelIndex index = m_treeView->currentIndex();
    QWidget* editor = m_treeView->indexWidget(index);
    if (editor) {
        // m_delegate->commitData(editor);
        m_delegate->closeEditor(editor, QAbstractItemDelegate::NoHint);
    }
    m_treeView->selectionModel()->clearSelection();
}

void ParameterTuningWidget::updateJobStatus()
{
    m_cautionSign->clear();

    if (jobItem()->isFailed()) {
        QString message;
        message.append("Current parameter values cause simulation failure.\n\n");
        message.append(jobItem()->comments());
        m_cautionSign->setCautionMessage(message);
    }

    updateDragAndDropSettings();
}
